function del_empty_lines()
  editor:BeginUndoAction()
  -- remember last caret line position
  local pre_l = editor:LineFromPosition(editor.CurrentPos)
  local ln = 0
  while ln < editor.LineCount do
    local text = editor:GetLine(ln)
    if not text then break end -- empty last line without EOL
    if string.match(text, "^%s*\r?\n$") then
      local p = editor:PositionFromLine(ln)
      editor:GotoPos(p)
      editor:LineDelete()
      -- adjust caret position if necessary
      if ln < pre_l then pre_l = pre_l - 1 end
    else
      -- move on if no more empty lines at this line number
      ln = ln + 1
    end
  end
  -- restore last caret line position
  local pre_p = editor:PositionFromLine(pre_l)
  editor:GotoPos(pre_p)
  editor:EndUndoAction()
end

function InsertDate()
	editor:AddText(os.date("%A, %m/%d/%Y"))
end


function CountOccurances()
  if props.CurrentSelection ~= "" then
	count = 0
    for m in editor:match( "^.*" .. props.CurrentSelection .. ".*$", SCFIND_REGEXP, 0) do
      count = count + 1
    end
	print(count)
  end
end


function WordCount()
	local whiteSpace = 0;   --number of whitespace chars
	local nonEmptyLine = 0; --number of non blank lines
	local wordCount = 0;    --total number of words
	
	--Calculate whitespace control
	for m in editor:match("\n") do
		whiteSpace = whiteSpace + 1;
	end
	for m in editor:match("\r") do
		whiteSpace = whiteSpace + 1;
	end
	for m in editor:match("\t") do --count tabs
		whiteSpace = whiteSpace + 1;
	end

	--Calculate non-empty lines and word count
	local itt = 0;
	while itt < editor.LineCount do --iterate through each line
		local hasChar, hasNum = 0;
		line = editor:GetLine(itt);
		if line then
			hasAlphaNum = string.find(line,'%w');
		end
		
		if (hasAlphaNum ~= nill) then
			nonEmptyLine = nonEmptyLine + 1;
		end
		
		if line then
			for word in string.gmatch(line, "%w+") do wordCount = wordCount + 1 end
		end
		
		itt = itt + 1;
	end
	
	print("----------------------------");
	print("Chars: \t\t\t",(editor.Length) - whiteSpace);
	print("Words: \t\t\t",wordCount);
	print("Lines: \t\t\t",editor.LineCount);
	print("Lines(non-blank): ", nonEmptyLine);
end

function stripTrailingSpaces(reportNoMatch)
    local count = 0
    local fs,fe = editor:findtext("[ \\t]+$", SCFIND_REGEXP)
    if fe then
        repeat
            count = count + 1
            editor:remove(fs,fe)
            fs,fe = editor:findtext("[ \\t]+$", SCFIND_REGEXP, fs)
        until not fe
        print("Removed trailing spaces from " .. count .. " line(s).")
    elseif reportNoMatch then
        print("Document was clean already; nothing to do.")
    end
    return count
end

function fixIndentation(reportNoMatch)
    local tabWidth = editor.TabWidth
    local count = 0
    if editor.UseTabs then
        -- for each piece of indentation that includes at least one space
        for m in editor:match("^[\\t ]* [\\t ]*", SCFIND_REGEXP) do
            -- figure out the indentation size
            local indentSize = editor.LineIndentation[editor:LineFromPosition(m.pos)]
            local spaceCount = math.mod(indentSize, tabWidth)
            local tabCount = (indentSize - spaceCount) / tabWidth
            local fixedIndentation = string.rep('\t', tabCount) .. string.rep(' ', spaceCount)

            if fixedIndentation ~= m.text then
                m:replace(fixedIndentation)
                count = count + 1
            end
        end
    else
        -- for each piece of indentation that includes at least one tab
        for m in editor:match("^[\\t ]*\t[\\t ]*", SCFIND_REGEXP) do
            -- just change all of the indentation to spaces
            m:replace(string.rep(' ', editor.LineIndentation[editor:LineFromPosition(m.pos)]))
            count = count + 1
        end
    end
    if count > 0 then
        print("Fixed indentation for " .. count .. " line(s).")
    elseif reportNoMatch then
        print("Document was clean already; nothing to do.")
    end
    return count
end

function lines(str)
  local t = {}
  local i, lstr = 1, #str
  while i <= lstr do
    local x, y = string.find(str, "\r?\n", i)
    if x then t[#t + 1] = string.sub(str, i, x - 1)
    else break
    end
    i = y + 1
  end
  if i <= lstr then t[#t + 1] = string.sub(str, i) end
  return t
end

function sort_text()
  local sel = editor:GetSelText()
  if #sel == 0 then return end
  local eol = string.match(sel, "\n$")
  local buf = lines(sel)
  --table.foreach(buf, print) --used for debugging
  table.sort(buf)
  local out = table.concat(buf, "\n")
  if eol then out = out.."\n" end
  editor:ReplaceSel(out)
end

function open_url()
  local string = string
  local function charat(s, p) return string.sub(s, p, p) end
  local function delim(c) return string.match(c, "[\"'%s]") end
  -- if there is a selection, use exactly, else analyze
  local txt = editor:GetSelText()
  if #txt == 0 then
    -- get details of current line, position
    local p1 = editor.CurrentPos
    local ln = editor:LineFromPosition(p1)
    txt = editor:GetLine(ln)
    if not txt then return end
    local p2 = editor:PositionFromLine(ln)
    p1 = p1 - p2 + 1; p2 = p1
    -- extend text segment to left
    while p1 > 1 do
      if delim(charat(txt, p1 - 1)) then break end
      p1 = p1 - 1
    end
    -- extend text segment to right
    while p2 <= #txt do
      if delim(charat(txt, p2)) then break end
      p2 = p2 + 1
    end
    -- exit if nothing matched
    if p1 == p2 then return end
    txt = string.sub(txt, p1, p2 - 1)
  else
    -- trim extraneous whitespace
    txt = string.gsub(txt, "^%s*(.-)%s*$", "%1")
    -- fail on embedded whitespace
    if string.match(txt, "%s") then return end
  end
  if string.match(txt, "^http://.+") or
     string.match(txt, "^ftp://.+") or
     string.match(txt, "^www%..+") then
    --print("URL='"..txt.."'") --DEBUG
    os.execute("runBrowser "..txt.." &")
  end
end

function clearOccurrences()
    scite.SendEditor(SCI_SETINDICATORCURRENT, 0)
    scite.SendEditor(SCI_INDICATORCLEARRANGE, 0, editor.Length)
end

function markOccurrences()
    if editor.SelectionStart == editor.SelectionEnd then
		os.execute("Xdialog --timeout 4 --title \"No Selection\" --msgbox \"\nYou need to make a selection, \nto mark occurrences for...\n\" 0 0")
        return
    end
    clearOccurrences()
    scite.SendEditor(SCI_INDICSETSTYLE, 0, INDIC_TT)
    scite.SendEditor(SCI_INDICSETFORE, 0, 255)
    local txt = GetCurrentWord()
    local flags = SCFIND_WHOLEWORD
    local s,e = editor:findtext(txt,flags,0)
    while s do
        scite.SendEditor(SCI_INDICATORFILLRANGE, s, e - s)
        s,e = editor:findtext(txt,flags,e+1)
    end
end

function isWordChar(char)
    local strChar = string.char(char)
    local beginIndex = string.find(strChar, '%w')
    if beginIndex ~= nil then
        return true
    end
    if strChar == '_' or strChar == '$' then
        return true
    end
    
    return false
end

function GetCurrentWord()
    local beginPos = editor.CurrentPos
    local endPos = beginPos
    if editor.SelectionStart ~= editor.SelectionEnd then
        return editor:GetSelText()
    end
    while isWordChar(editor.CharAt[beginPos-1]) do
        beginPos = beginPos - 1
    end
    while isWordChar(editor.CharAt[endPos]) do
        endPos = endPos + 1
    end
    return editor:textrange(beginPos,endPos)
end

function caseToggle(reportNoMatch)
  local selText = editor:GetSelText()
  if selText ~= ""  and  selText ~= nil then
	 local selFrom = editor.SelectionStart
	 local selTo = editor.SelectionEnd
	 if selText == string.lower(selText) then
		editor:ReplaceSel(string.upper(selText))
	 else
		editor:ReplaceSel(string.lower(selText))
	 end
	 editor:SetSel(selFrom, selTo)
  else
	 print ("caseToggle:: Must select text first.")
  end
  return
end -- caseToggle


function wrap_text()

    local border = 80
    local t = {}

    local pos = editor.SelectionStart
    local sel = editor:GetSelText()
    if #sel == 0 then return end

    local para = {}
    local function helper(line) table.insert(para, line) return "" end
    helper((sel:gsub("(.-)\r?\n", helper)))

    for k, v in pairs(para) do
        line = ""
        for token in string.gmatch(v, "[^%s]+") do
            if string.len(token .. line) >= border then
                t[#t + 1] = line
                line = token .. " "
            else
                line = line .. token .. " "
            end
        end
        t[#t + 1] = line:gsub("%s$", "")
    end

    editor:ReplaceSel(table.concat(t, "\n"))
    editor:GotoPos(pos)

end

function html_entities()
local entities = {
[' '] = '&nbsp;' ,
[''] = '&iexcl;' ,
[''] = '&cent;' ,
[''] = '&pound;' ,
[''] = '&curren;' ,
[''] = '&yen;' ,
[''] = '&brvbar;' ,
[''] = '&sect;' ,
[''] = '&uml;' ,
[''] = '&copy;' ,
[''] = '&ordf;' ,
[''] = '&laquo;' ,
[''] = '&not;' ,
[''] = '&shy;' ,
[''] = '&reg;' ,
[''] = '&macr;' ,
[''] = '&deg;' ,
[''] = '&plusmn;' ,
[''] = '&sup2;' ,
[''] = '&sup3;' ,
[''] = '&acute;' ,
[''] = '&micro;' ,
[''] = '&para;' ,
[''] = '&middot;' ,
[''] = '&cedil;' ,
[''] = '&sup1;' ,
[''] = '&ordm;' ,
[''] = '&raquo;' ,
[''] = '&frac14;' ,
[''] = '&frac12;' ,
[''] = '&frac34;' ,
[''] = '&iquest;' ,
[''] = '&Agrave;' ,
[''] = '&Aacute;' ,
[''] = '&Acirc;' ,
[''] = '&Atilde;' ,
[''] = '&Auml;' ,
[''] = '&Aring;' ,
[''] = '&AElig;' ,
[''] = '&Ccedil;' ,
[''] = '&Egrave;' ,
[''] = '&Eacute;' ,
[''] = '&Ecirc;' ,
[''] = '&Euml;' ,
[''] = '&Igrave;' ,
[''] = '&Iacute;' ,
[''] = '&Icirc;' ,
[''] = '&Iuml;' ,
[''] = '&ETH;' ,
[''] = '&Ntilde;' ,
[''] = '&Ograve;' ,
[''] = '&Oacute;' ,
[''] = '&Ocirc;' ,
[''] = '&Otilde;' ,
[''] = '&Ouml;' ,
[''] = '&times;' ,
[''] = '&Oslash;' ,
[''] = '&Ugrave;' ,
[''] = '&Uacute;' ,
[''] = '&Ucirc;' ,
[''] = '&Uuml;' ,
[''] = '&Yacute;' ,
[''] = '&THORN;' ,
[''] = '&szlig;' ,
[''] = '&agrave;' ,
[''] = '&aacute;' ,
[''] = '&acirc;' ,
[''] = '&atilde;' ,
[''] = '&auml;' ,
[''] = '&aring;' ,
[''] = '&aelig;' ,
[''] = '&ccedil;' ,
[''] = '&egrave;' ,
[''] = '&eacute;' ,
[''] = '&ecirc;' ,
[''] = '&euml;' ,
[''] = '&igrave;' ,
[''] = '&iacute;' ,
[''] = '&icirc;' ,
[''] = '&iuml;' ,
[''] = '&eth;' ,
[''] = '&ntilde;' ,
[''] = '&ograve;' ,
[''] = '&oacute;' ,
[''] = '&ocirc;' ,
[''] = '&otilde;' ,
[''] = '&ouml;' ,
[''] = '&divide;' ,
[''] = '&oslash;' ,
[''] = '&ugrave;' ,
[''] = '&uacute;' ,
[''] = '&ucirc;' ,
[''] = '&uuml;' ,
[''] = '&yacute;' ,
[''] = '&thorn;' ,
[''] = '&yuml;' ,
['"'] = '&quot;' ,
["'"] = '&#39;' ,
['<'] = '&lt;' ,
['>'] = '&gt;' ,
['&'] = '&amp;'
}
  -- Here we search for non standar characters and replace them if
  -- we have a translation. The regexp could be changed to include an
  -- exact list with the above characters [...] and then remove
  -- the 'if' below, but it's easier to maintain like this...
  editor:ReplaceSel(string.gsub(editor:GetSelText(), "[^a-zA-Z0-9 _]",
    function (v)
      if entities[v] then return entities[v] else return v end
    end)
  )
end
function OnClear()
    scite.StripShow("")
end

function spell()
    spell_table = {}
    offset = 0
    local ln = 0
    local text = editor:GetSelText()
    if text == "" then text = (editor:GetText()) else
        ln = editor:LineFromPosition(editor.SelectionStart)
        if editor:PositionFromLine(ln) ~= editor.SelectionStart then
            offset = editor.SelectionStart - editor:PositionFromLine(ln) end
    end
    editor.IndicatorCurrent = 2
    editor.IndicStyle[2] = INDIC_SQUIGGLE
    local h_out = io.popen('sh -c "sed \'s/^$/ /g\' | aspell -a" << p-i-p-e\n' .. text:gsub('`', ' ') .. '\np-i-p-e')
    for l in h_out:lines() do
        if l == "" then ln = ln + 1 end
        for k, v in string.gmatch(l, "[&|#] (.+): (.+)") do
            local w = k:match("(%w+)")
            local c = k:match(" (%w+)$")
            if ln == editor:LineFromPosition(editor.SelectionStart) then c = c + offset end
            editor:GotoLine(ln)
            editor:IndicatorFillRange(editor.CurrentPos + c, w:len())
            table.insert(spell_table, {line = ln, column = c, word = w, suggestions = v})
        end
    end
    if #spell_table == 0 then scite.StripShow("'Spell-check completed: no misspelled words'") else
        h_out:close()
        scite.StripShow("'Current Word:'[](&Add)(&Skip)\n'Suggestion:'{}(&Replace)(&Cancel)")
        process_spell_matches()
    end
end

function clear_field()
    spell_table = nil
    editor:SetSel(0, 0)
    editor.IndicatorCurrent = 2
    editor:IndicatorClearRange(0, editor.TextLength)
    scite.StripShow("")
end

function process_spell_matches(x)
    local h = spell_table[1]
    if h then
        if y ~= h.line then offset = 0 else
            if x then offset = offset + x - wl end
        end
        y = h.line
        table.remove(spell_table, 1)
        editor:GotoLine(y)
        local pos = editor.CurrentPos + h.column
        editor:SetSel(pos + offset, pos + offset + h.word:len())
        scite.StripSet(1, h.word)
        scite.StripSet(5, h.suggestions:match("(%w+)"))
        scite.StripSetList(5, "")
        scite.StripSetList(5, h.suggestions:gsub(", ", "\n"))
    else clear_field() end
end

function OnStrip(control, change)
    if change == 1 then
        if control == 2 then  -- Add
            local a = io.open(props["SciteUserHome"] .. "/.aspell.en.pws", "a")
            if a then a:write(scite.StripValue(1) .. "\n") end
            a:close()
            process_spell_matches()
        end
        if control == 3 then process_spell_matches() end  -- Next
        if control == 6 then  -- Replace
            _, wl = editor:GetSelText()
            if wl > 0 then wl = wl - 1; editor:ReplaceSel(scite.StripValue(5)) end
            process_spell_matches(scite.StripValue(5):len())
        end
        if control == 7 then clear_field() end  -- Cancel
    end
end